#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>
#include <assert.h>
#include <strings.h>


void Error(char *message)
{
    printf("%s\n", message);
    exit(0);
}


FILE *File_Open(char *filename, char *mode)
{
    FILE *f = fopen(filename, mode);

    if (f == NULL) Error("Cannot open file");
    
    return f;
}


void File_Close(FILE *f)
{
    if (fclose(f) != 0) Error("Cannot close file");
}


void File_Write(FILE *f, void *buf, uint32_t size)
{
    if (fwrite(buf, 1, size, f) != size) Error("File write error");
}


void File_Read(FILE *f, void *buf, uint32_t size)
{
    if (fread(buf, 1, size, f) != size) Error("File read error");
}


void File_WriteInt8(FILE *f, uint8_t n)
{
    File_Write(f, &n, sizeof(n));
}


void File_WriteInt16LE(FILE *f, uint16_t n)
{
    uint8_t le[2] = {(uint8_t)(n >> 0), (uint8_t)(n >> 8)};
    File_Write(f, le, sizeof(le));
}


void File_WriteInt32LE(FILE *f, uint32_t n)
{
    uint8_t le[4] = {(uint8_t)(n >> 0), (uint8_t)(n >> 8), (uint8_t)(n >> 16), (uint8_t)(n >> 24)};
    File_Write(f, le, sizeof(le));
}



uint16_t File_ReadInt16BE(FILE *f)
{
    uint8_t be[2];

    File_Read(f, be, sizeof(be));

    return ((uint16_t)be[0] << 8) | ((uint16_t)be[1] << 0);
}


uint32_t File_ReadInt32BE(FILE *f)
{
    uint8_t be[4];

    File_Read(f, be, sizeof(be));

    return ((uint32_t)be[0] << 24) | ((uint32_t)be[1] << 16) | ((uint32_t)be[2] << 8) | ((uint32_t)be[3] << 0);
}


uint32_t File_GetSize(FILE *f)
{
    long pos;
    uint32_t size;
    
    assert((pos = ftell(f)) != -1);

    assert(fseek(f, 0, SEEK_END) == 0);
    assert((size = (unsigned)ftell(f)) != (unsigned)-1);

    assert(fseek(f, pos, SEEK_SET) == 0);

    return size;
}


void File_Seek(FILE *f, uint32_t pos)
{
    if (fseek(f, pos, SEEK_SET) != 0) Error("File seek error");
}


void *Mem_Alloc(uint32_t size)
{
    void *buf;
    
    if ((buf = malloc(size)) == NULL) Error("Out of memory");
    
    return buf;
}


void Mem_Free(void *buf)
{
    free(buf);
}


int main(int argc, char *argv[])
{
    FILE *f_img, *f_pal, *f_out;
    uint32_t width, height;
    uint32_t bmplinesize, bmpfilesize;
    uint8_t *linebuffer;
    int32_t i;

    if (argc != 5) Error("Usage: 8bit_to_bmp <image file> <palette file> <output file> <image width>");

    if (sscanf(argv[4], "%u", &width) != 1) Error("Invalid image width");
    if (width == 0) Error("Image width cannot be zero");

    f_img = File_Open(argv[1], "rb");
    f_pal = File_Open(argv[2], "rb");
    f_out = File_Open(argv[3], "wb");
    
    height = File_GetSize(f_img) / width;

    bmplinesize = width;
    bmplinesize = (bmplinesize + 3) & (~3);
    bmpfilesize = 54 + (256 * 4) + (bmplinesize * height);
    
    File_WriteInt16LE(f_out, 0x4D42);           /* bfType */
    File_WriteInt32LE(f_out, bmpfilesize);      /* bfSize */
    File_WriteInt16LE(f_out, 0);                /* bfReserved1 */
    File_WriteInt16LE(f_out, 0);                /* bfReserved2 */
    File_WriteInt32LE(f_out, 54 + (256 * 4));   /* bfOffBits */

    File_WriteInt32LE(f_out, 40);               /* biSize */
    File_WriteInt32LE(f_out, width);            /* biWidth */
    File_WriteInt32LE(f_out, height);           /* biHeight */
    File_WriteInt16LE(f_out, 1);                /* biPlanes */
    File_WriteInt16LE(f_out, 8);                /* biBitCount */
    File_WriteInt32LE(f_out, 0);                /* biCompression */
    File_WriteInt32LE(f_out, 0);                /* biSizeImage */
    File_WriteInt32LE(f_out, 0);                /* biXPelsPerMeter */
    File_WriteInt32LE(f_out, 0);                /* biYPelsPerMeter */
    File_WriteInt32LE(f_out, 0);                /* biClrUsed */
    File_WriteInt32LE(f_out, 0);                /* biClrImportant */

    if (File_GetSize(f_pal) < 512) Error("Bad palette file");
    for (i = 0; i < 256; i++)
    {
        uint16_t c = File_ReadInt16BE(f_pal);

        File_WriteInt8(f_out, ((c >> 6) & 0x1F) << 3);
        File_WriteInt8(f_out, ((c >> 0) & 0x3F) << 2);
        File_WriteInt8(f_out, ((c >> 11) & 0x1F) << 3);
        File_WriteInt8(f_out, 0);
    }

    linebuffer = Mem_Alloc(bmplinesize);
    memset(linebuffer, 0, bmplinesize);
    for (i = height - 1; i >= 0; i--)
    {
        File_Seek(f_img, i * width);
        File_Read(f_img, linebuffer, width);
        File_Write(f_out, linebuffer, bmplinesize);
    }
    Mem_Free(linebuffer);

    File_Close(f_img);
    File_Close(f_pal);
    File_Close(f_out);

    return 0;
}
